import json
from copy import deepcopy
from typing import Union
from datetime import datetime

class AnnotationCreator:

    def __init__(self, data:Union[str,dict]=None):
        self.__data__ = {}
        self.__data__['applyToAllReports'] = False
        self.__data__['scope'] = {'metrics': [], 'filters': []}
        self.__data__['tags'] = []
        if isinstance(data, str):
            if data.endswith('.json'):
                with open(data, 'r') as file:
                    self.__data__ = json.load(file)
        elif isinstance(data, dict):
            self.__data__ = deepcopy(data)

    def __str__(self):
        return json.dumps(self.__data__, indent=4)
    
    def __repr__(self):
        return json.dumps(self.__data__, indent=2)
    
    def to_dict(self):
        """
        Convert the annotation data to a dictionary.
        Returns:
            dict: The annotation data as a dictionary.
        """
        return deepcopy(self.__data__)


    def setName(self, name: str):
        """
        Set the name for the annotation.
        Arguments:
            name: REQUIRED : Name to be used for the annotation.
        """
        if not name:
            raise ValueError("Name cannot be empty")
        self.__data__['name'] = name
    
    def setDescription(self, description: str):
        """
        Set the description for the annotation.
        Arguments:
            description: REQUIRED : Description to be used for the annotation.
        """
        self.__data__['description'] = description
    
    def setColor(self, color: str):
        """
        Set the color for the annotation.
        Arguments:
            color: Color to be used, from "STANDARD1" to "STANDARD9"
        """
        if color not in ["STANDARD1", "STANDARD2", "STANDARD3", "STANDARD4", "STANDARD5","STANDARD6","STANDARD7","STANDARD8","STANDARD9"]:
            raise ValueError("Invalid color. Choose from STANDARD1, STANDARD2, STANDARD3, STANDARD4, STANDARD5,STANDARD6,STANDARD7,STANDARD8,STANDARD9")
        self.__data__['color'] = color

    def setReportSuiteName(self, reportSuiteName: str):
        """
        Set the report suite Name for the annotation.
        Arguments:
            reportSuiteName: REQUIRED : Report suite Name to be used.
        """
        self.__data__['reportSuiteName'] = reportSuiteName

    def setReportSuiteId(self, reportSuiteId: str):
        """
        Set the report suite ID for the annotation.
        Arguments:
            reportSuiteId: REQUIRED : Report suite ID to be used.
        """
        self.__data__['reportSuiteId'] = reportSuiteId
    
    def applyToAllReports(self, apply: bool):
        """
        Set whether the annotation applies to all reports.
        Arguments:
            apply: REQUIRED : Boolean indicating if the annotation applies to all reports.
        """
        self.__data__['applyToAllReports'] = apply

    def setDateRange(self, dateRange: str = None,start: str=None, end: str=None):
        """
        Set the date range for the annotation.
        There are two ways to set the date range:
        1. Using a direct date range in ISO format (YYYY-MM-DDT00:00:00/YYYY-MM-DDT00:00:00).
        2. Using start and end dates in ISO format (YYYY-MM-DD). The wrapper will create the date range automatically.
        if both are provided, the dateRange will be used.
        Arguments:
            dateRange: OPTIONAL : Direct date range in ISO format (YYYY-MM-DDT00:00:00/YYYY-MM-DDT00:00:00).
            start: OPTIONAL : Start date in ISO format (YYYY-MM-DD).
            end: OPTIONAL : End date in ISO format (YYYY-MM-DD).
        """
        if dateRange:
            self.__data__['dateRange'] = dateRange
        else:
            if start is None or end is None:
                raise ValueError("Either dateRange or both start and end must be provided")
            start = datetime.fromisoformat(start)
            end = datetime.fromisoformat(end)
            if start > end:
                raise ValueError("Start date cannot be after end date")
            self.__data__['dateRange'] = f"{start.isoformat()}/{end.isoformat()}"

    def addMetric(self, metricId: str):
        """
        Add a single metric to the annotation.
        Arguments:
            metricId: REQUIRED : ID of the metric to be added.
        """
        if 'scope' not in self.__data__.keys():
            self.__data__['scope'] = {'metrics': [],'filters': []}
        self.__data__['scope']['metrics'].append({"id": metricId,"componentType": "metric"})
    
    def addMetrics(self, metricIds: list):
        """
        Add multiple metrics to the annotation.
        Arguments:
            metricIds: REQUIRED : List of metric IDs to be added.
        """
        if not isinstance(metricIds, list):
            raise TypeError("metricIds must be a list")
        if 'scope' not in self.__data__.keys():
            self.__data__['scope'] = {'metrics': [],"filters": []}
        for metricId in metricIds:
            self.__data__['scope']['metrics'].append({"id": metricId,"componentType": "metric"})

    def addFilter(self, filterId: str,operator:str=None,dimensionType:str=None,terms:list=None,componentType:str=None):
        """
        Add a filter to the annotation.
        Arguments:
            filterId: ID of the filter to be added.
            operator: Optional operator for the filter.
            dimensionType: Optional dimension type for the filter.
            terms: Optional list of terms for the filter.
            componentType: Optional component type for the filter.
        """
        if 'scope' not in self.__data__.keys():
            self.__data__['scope'] = {'metrics': [],"filters": []}
        if 'filters' not in self.__data__['scope']:
            self.__data__['scope']['filters'] = []
        filter_data = {"id": filterId}
        if operator:
            filter_data['operator'] = operator
        if dimensionType:
            filter_data['dimensionType'] = dimensionType
        if terms:
            filter_data['terms'] = terms
        if componentType:
            filter_data['componentType'] = componentType
        elif filterId.startswith('s'):
            filter_data['componentType'] = 'segment'
        self.__data__['scope']['filters'].append(filter_data)
    
    def addTag(self, tagId: str,name:str=None):
        """
        Add a tag to the annotation.
        Arguments:
            tag: REQUIRED : Tag to be added.
            name: REQUIRED : name for the tag.
        """
        if tagId is None or tagId == "":
            raise ValueError("Tag ID cannot be None or empty")
        if name is None or name == "":
            raise ValueError("Tag name cannot be None or empty")
        if 'tags' not in self.__data__.keys():
            self.__data__['tags'] = []
        self.__data__['tags'].append({tagId: tagId, "name": name})

    



# Non standard libraries
import pandas as pd
from aanalytics2 import config, connector
from typing import Union
from copy import deepcopy
import pandas as pd
import json
import time

class Report:

    def __init__(self,df:pd.DataFrame,reportType:str,rsid:str,period:str)->None:
        """
        Build a class to return the report with additional information.
        Arguments:
            df : REQUIRED : The dataframe build out of the report
            reportType : REQUIRED : The type of report used
            rsid : REQUIRED : The reportSuite ID used
            period : OPTIONAL : if a ranked report        
        """
        self.DataFrame = df.copy()
        self.columns = df.columns.to_list()
        self.reportType = reportType
        self.rsid = rsid
        if period is not None:
            self.period = period

    def __str__(self):
        """Returns the definition"""
        return f"""Report from {self.rsid}. Type: {self.reportType}. Columns:{self.columns} """
    
    def __repr__(self):
        return self.DataFrame.to_string()
    
    def getDataFrame(self):
        return self.DataFrame

class LegacyAnalytics:
    """
    Class that will help you realize basic requests to the old API 1.4 endpoints
    """

    def __init__(self,company_name:str=None,config:dict=config.config_object)->None:
        """
        Instancialize the Legacy Analytics wrapper.
        """
        if company_name is None:
            raise Exception("Require a company name")
        self.connector = connector.AdobeRequest(config_object=config)
        self.token = self.connector.token
        self.endpoint = "https://api.omniture.com/admin/1.4/rest"
        self.header = header = {
            'Accept': 'application/json',
            'Authorization': f'Bearer {self.token}',
            'X-ADOBE-DMA-COMPANY': company_name
        }

    
    def getData(self,path:str="/",method:str=None,params:dict=None)->dict:
        """
        Use the GET method to the parameter used.
        Arguments:
            path : REQUIRED : If you need a specific path (default "/")
            method : OPTIONAL : if you want to pass the method directly there for the parameter.
            params : OPTIONAL : If you need to pass parameter to your url, use dictionary i.e. : {"param":"value"} 
        """
        if params is not None and type(params) != dict:
            raise TypeError("Require a dictionary")
        myParams = {}
        myParams.update(**params or {})
        if method is not None:
            myParams['method'] = method
        path = path
        res = self.connector.getData(self.endpoint + path,params=myParams,headers=self.header,legacy=True)
        return res

    def postData(self,path:str="/",method:str=None,params:dict=None,data:Union[dict,list]=None)->dict:
        """
        Use the POST method to the parameter used.
        Arguments:
            path : REQUIRED : If you need a specific path (default "/")
            method : OPTIONAL : if you want to pass the method directly there for the parameter.
            params : OPTIONAL : If you need to pass parameter to your url, use dictionary i.e. : {"param":"value"}
            data : OPTIONAL : Usually required to pass the dictionary or list to the request
        """
        if params is not None and type(params) != dict:
            raise TypeError("Require a dictionary")
        if data is not None and (type(data) != dict and type(data) != list):
            raise TypeError("data should be dictionary or list")
        myParams = {}
        myParams.update(**params or {})
        if method is not None:
            myParams['method'] = method
        path = path
        res = self.connector.postData(self.endpoint + path,params=myParams, data=data,headers=self.header,legacy=True)
        return res

    def getReportWait(self,reportDescription:Union[dict,'ReportBuilder14']=None,wait:int=60)->pd.DataFrame:
        """
        Take a report description, send it to the queue and wait for the data in a loop.
        Returns a dataframe
        Argument:
            reportDescription : REQUIRED : The data report definition needed (dictionary or a ReportBuilder instance)
            wait : OPTIONAL : How long in seconds to wait before next request to get data
        """
        if reportDescription is None:
            raise Exception("Require a reportDescription")
        if hasattr(reportDescription,'to_dict'):
            reportDescription = reportDescription.to_dict()
        methodQueue = "Report.Queue"
        res = self.postData(method=methodQueue,data=reportDescription)
        if 'reportID' not in res.keys():
            raise Exception(res)
        data = pd.DataFrame
        while data.empty:
            methodGet = "Report.Get"
            mydata = self.postData(method=methodGet,data=res)
            if 'report' in mydata.keys():
                data = self.transformReportToDataFrame(mydata)
                return data
            time.sleep(wait)


    def __recursiveBuild__(self,row,context:list=None,level:int=0,targetLevel:int=0) -> list:
        """
        Recursive function to build the pandas data frame.
        Arguments:
            row : row of data contain in hte ['report']['data'] object.
            context : list of elements to append to each row
            level : level at which the recursion is
            targetLevel : level expected to reach the recursion.
        """
        level +=1
        if type(row) == dict:
            if context is None:
                context = [row['name']]
            else:
                context.append(row['name'])
            if 'breakdown' in row.keys():   
                data = self.__recursiveBuild__(row['breakdown'],context,level,targetLevel)
            else:
                data = []
                if level != targetLevel: ## if level is different it means we have empty rows and missing elements. Feeling with none
                    new_row = context + [None]*(targetLevel-level) + row["counts"]
                else:
                    new_row = [context  + row["counts"]]
                data += new_row
        elif type(row) == list:
            data = []
            for r in row:
                if 'breakdown' in r.keys():
                    row_context = deepcopy(context)
                    if row_context is None:
                        row_context = [r['name']]
                    else:
                        row_context.append(r['name'])
                    data += self.__recursiveBuild__(r['breakdown'],row_context,level,targetLevel)
                else:
                    row_context = deepcopy(context)
                    if level != targetLevel:
                        new_row = row_context + [r["name"]] + [None]*(targetLevel-level) + r["counts"]
                    else:
                        if row_context is None:
                            new_row = [r["name"]] + r["counts"]
                        else:
                            new_row = row_context + [r["name"]] + r["counts"]
                    data.append(new_row)
        return data
    
    def transformReportToDataFrame(self,data:dict,reportClass:bool=False)->pd.DataFrame:
        """
        Function to pass the response from the "Report.Get" methods when data has been received.
        It will output a dataframe
        Arguments:
            data : REQUIRED : The data response from the request "Report.Get".
            reportClass : OPTIONAL : If you want more context of the data.
        """
        if data is None:
            raise ValueError('Require data to be passed')
        reportType = data.get("report",{}).get("type")
        if reportType is not None:
            dimNames = [el['name'] if 'classification' not in el.keys() else el['classification'] for el in data.get("report",{}).get("elements",[])]
            metrics = [el['name'] for el in data.get("report",{}).get("metrics",[])]
            if reportType=="trended":
                cols = ['date']+dimNames+metrics
                levelsTarget = len(dimNames) +1
            else:
                cols = dimNames+metrics
                levelsTarget = len(dimNames)
            results = []
            for row in data['report']['data']:
                results += self.__recursiveBuild__(row,targetLevel=levelsTarget)
            df = pd.DataFrame(results)
            df.columns = cols
            if reportClass:
                class_data = Report(df,
                                    reportType,
                                    data.get('report').get('reportSuite').get('id'),
                                    data.get('report').get('period'))
                return class_data
            return df

class ReportBuilder14:

    """
    Build a report for the Legacy Analytics 1.4 API Report capability. 
    """

    def __init__(self,template:Union[dict,'ReportBuilder14']=None)->None:
        """
        Construct a report based information.
        Arguments:
            template : OPTIONAL : If you want to use a template of a report to not start from scratch.
        """
        report = {'reportDescription':{'reportSuiteID':None,'elements':[],'metrics':[]}}
        if template is not None:
            if type(template) == ReportBuilder14:
                report = template.to_dict()
            elif type(template) == dict:
                if 'reportDescription' in template.keys():
                    template = template['reportDescription']
                for key in template.keys():
                    report['reportDescription'][key] = template[key]
                if 'metrics' not in report['reportDescription'].keys():
                    report['metrics'] = []
                if 'elements' not in report['reportDescription'].keys():
                    report['elements'] = []
        self.report = report
    
    def __str__(self):
        return json.dumps(self.report,indent=2)
    
    def __repr__(self):
        return json.dumps(self.report,indent=2)
    
    def to_dict(self):
        """
        Return the report as dictionary
        """
        return deepcopy(self.report)

    def setReportSuite(self,rsid:str):
        """
        Add a reportSuite
        Argument:
            rside : REQUIRED : The reportSuite ID
        """
        self.report["reportDescription"]['reportSuiteID'] = rsid
    
    def setSource(self,source:str="warehouse"):
        """
        In case yu want to define data warehouse report per example
        Argument: 
            source : REQUIRED : The source, by default "warehouse"
        """
        self.report["reportDescription"]["source"] = source
    
    def removeSource(self):
        """
        remove the source key in the report.
        """
        if 'source' in self.report['reportDescription'].keys():
            del self.report['reportDescription']['source']
    
    def setDate(self,date:str=None,dateFrom:str=None,dateTo:str=None):
        """
        Set the date for the report. You can use either date or the dateFrom and dateTo parameter.
        Arguments:
            date : OPTIONAL : If requesting a report for a single period, the day/month/year that you want to run the report for. If you use this field, do not use dateFrom or dateTo. If all date fields are omitted, defaults to the current day. This field supports the following formats:
                    Use YYYY for the desired year.
                    Use YYYY-MM for the desired month.
                    Use YYYY-MM-DD for the desired day.
            dateFrom : OPTIONAL : The starting date range. If you use this field, also include dateTo and do not use date. The date format is YYYY-MM-DD. The month and day designators can be omitted if you want a monthly or yearly report
            dateTo : OPTIONAL : The ending date range (inclusive). If you use this field, also include dateFrom and do not use date. The date format is YYYY-MM-DD. The month and day designators can be omitted if you want a monthly or yearly report
        """
        
        if date is not None:
            if 'dateTo' in self.report['reportDescription'].keys():
                del self.report['reportDescription']["dateTo"]
            if 'dateFrom' in self.report['reportDescription'].keys():
                del self.report['reportDescription']["dateFrom"]
            self.report['reportDescription']['date'] = date
        else:
            if dateTo is None or dateFrom is None:
                raise Exception("Require a date parameter to be used")
            else:
                if "date" in self.report['reportDescription'].keys():
                    del self.report['reportDescription']["date"]
                self.report['reportDescription']['dateFrom'] = dateFrom
                self.report['reportDescription']['dateTo'] = dateTo

    def setGranularity(self,dateGranularity:str=None):
        """
        Specifies the date granularity used to display report data (trended reports). If this field is omitted, data is aggregated across the entire date range (ranked reports).
        Argument:
            dateGranularity : REQUIRED : Supported values are:
                minute: Real-Time reports only. Displays report data by minute. Include an integer between 1 - 60 after a semicolon to increase the minute interval. For example, use minute:3 to aggregate data in 3-minute intervals.
                hour: Displays report data by hour.
                day: Displays report data by day.
                week: Displays report data by week.
                month: Displays report data by month.
                quarter: Displays report data by quarter.
                year: Displays report data by year.
        """
        self.report['reportDescription']['dateGranularity'] = dateGranularity

    def removeGranularity(self):
        """
        Remove the granularity component
        """
        if 'dateGranularity' in self.report['reportDescription'].keys():
            del self.report['reportDescription']['dateGranularity']
    
    def addMetric(self,metricId:str=None,**kwargs):
        """
        Add a specific metric
        Argument: 
            metricId : REQUIRED : The metric ID to be used
            kwargs : OPTIONAL : any additional attribute you want to add
        """
        dict_metrics = {
            "id":metricId
        }
        if len(kwargs)>0:
            dict_metrics.update(**kwargs)
        self.report['reportDescription']['metrics'].append(dict_metrics)

    def addElement(self,elementId:str=None,**kwargs):
        """
        Add a specific metric
        Argument: 
            elementId : REQUIRED : The element ID to be used
            kwargs : OPTIONAL : any additional attribute you want to add
        """
        dict_element = {
            "id":elementId
        }
        if len(kwargs)>0:
            dict_element.update(**kwargs)
        self.report['reportDescription']['elements'].append(dict_element)


    def setEncoding(self,encoding:str="utf8"):
        """
        Supported values include utf8 or base64.
        Arguments:
            encodning : OPTIONAL : either utf-8 or bas64.
                utf8: Filters out invalid UTF-8 characters in the request and response.
                base64: Treats the entire request, including element names, search/pathing filters, special keywords, and dates, as if they are base64 encoded.
        """
        self.report['reportDescription']['elementDataEncoding'] = encoding

    def addSegment(self,segment:dict=None):
        """
        Require you to pass the segment object definition to add it on the request.
        Arguments
            segment : REQUIRED : A dictionary that provides some information on the segment depending the type.
        
        Example: 
            segment = {
                "id" : "segmentID"
            }
            or 
            segment = {
                "element": "page",
                "selected": ["Home Page", "Shopping Cart"]
            }
            or 
            segment = {
                "element": "eVar1",
                "classification": "Group Name",
                "search": { "type": "OR", "keywords": ["Administrator", "Manager", "Director"]
            }
        """
        if 'segments' not in self.report['reportDescription'].keys():
            self.report['reportDescription']['segments'] = []
        self.report['reportDescription']['segments'].append(segment)

    def removeSegment(self):
        """
        Remove the segment from the report request.
        """
        if 'segments' in self.report['reportDescription'].keys():
           del self.report['reportDescription']['segments']


import pytest

import aanalytics2


def test_should_raise_on_empty_path_and_empty_private_key():
    with pytest.raises(ValueError):
        aanalytics2.configure(org_id='org_id',
                              tech_id='tech_id',
                              secret='secret',
                              path_to_key='',
                              client_id='client_id')


def test_private_key_should_be_used_if_provided():
    expected_private_key = 'private_key'
    aanalytics2.configure(org_id='org_id',
                          tech_id='tech_id',
                          secret='secret',
                          path_to_key=None,
                          client_id='client_id',
                          private_key=expected_private_key)
    config = aanalytics2.config_object
    assert aanalytics2.configs.get_private_key_from_config(config) == expected_private_key



config_object = {
    "org_id": "",
    "client_id": "",
    "tech_id": None,
    "pathToKey": None,
    "secret": "",
    "scopes":None,
    "date_limit": 0,
    "token": "",
    "jwtTokenEndpoint": "https://ims-na1.adobelogin.com/ims/exchange/jwt",
    "oauthTokenEndpointV2" : "https://ims-na1.adobelogin.com/ims/token/v3"
}
orga_admin = {'_org_admin', '_deployment_admin', '_support_admin'}
header = {"Accept": "application/json",
          "Content-Type": "application/json",
          "Authorization": "Bearer ",
          "x-api-key": ""
          }
endpoints = {
    "global": 'https://analytics.adobe.io/api',
    "ingestion": "https://analytics-collection.adobe.io"
}



import json
import time
from copy import deepcopy

# Non standard libraries
import requests
import io
from aanalytics2 import config, token_provider


class AdobeRequest:
    """
    Handle request to Audience Manager and taking care that the request have a valid token set each time.
    Attributes:
        restTime : Time to rest before sending new request when reaching too many request status code.
    """
    loggingEnabled = False
    def __init__(self,
                 config_object: dict = config.config_object,
                 header: dict = config.header,
                 verbose: bool = False,
                 retry: int = 0,
                 loggingEnabled:bool=False,
                 logger:object=None,
                 company_id:str=None
                ) -> None:
        """
        Set the connector to be used for handling request to AAM
        Arguments:
            config_object : OPTIONAL : Require the importConfig file to have been used.
            header : OPTIONAL : header of the config modules
            verbose : OPTIONAL : display comment on the request.
            retry : OPTIONAL : If you wish to retry failed GET requests
            loggingEnabled : OPTIONAL : if the logging is enable for that instance.
            logger : OPTIONAL : instance of the logger created
            company_id: OPTIONAL : company id to be used for the request
        """
        if config_object['org_id'] == '':
            raise Exception(
                'You have to upload the configuration file with importConfigFile method.')
        self.config = deepcopy(config_object)
        self.header = deepcopy(header)
        self.loggingEnabled = loggingEnabled
        self.logger = logger
        self.restTime = 30
        self.retry = retry
        if self.config['token'] == '' or time.time() > self.config['date_limit']:
            if 'scopes' in self.config.keys() and self.config.get('scopes',None) is not None:
                self.connectionType = 'oauthV2'
                token_and_expiry = token_provider.get_oauth_token_and_expiry_for_config(config=self.config, verbose=verbose)
            elif self.config.get("private_key",None) is not None or self.config.get("pathToKey",None) is not None:
                self.connectionType = 'jwt'
                token_and_expiry = token_provider.get_jwt_token_and_expiry_for_config(config=self.config, verbose=verbose)
            token = token_and_expiry['token']
            expiry = token_and_expiry['expiry']
            self.token = token
            if self.loggingEnabled:
                self.logger.info("token retrieved : {token}")
            self.config['token'] = token
            self.config['date_limit'] = time.time() + expiry - 500
            self.header.update({'Authorization': f'Bearer {token}'})
            self.header.update({'x-proxy-global-company-id': company_id})

    def _checkingDate(self) -> None:
        """
        Checking if the token is still valid
        """
        now = time.time()
        if now > self.config['date_limit']:
            if self.loggingEnabled:
                self.logger.warning("token expired. Trying to retrieve a new token")
            if self.connectionType =='oauthV2':
                token_and_expiry = token_provider.get_oauth_token_and_expiry_for_config(config=self.config)
            elif self.connectionType == 'jwt':
                token_and_expiry = token_provider.get_jwt_token_and_expiry_for_config(config=self.config)
            token = token_and_expiry['token']
            if self.loggingEnabled:
                self.logger.info(f"new token retrieved : {token}")
            self.config['token'] = token
            self.config['date_limit'] = time.time() + token_and_expiry['expiry'] - 500
            self.header.update({'Authorization': f'Bearer {token}'})

    def getData(self, endpoint: str, params: dict = None, data: dict = None, headers: dict = None, *args, **kwargs):
        """
        Abstraction for getting data
        """
        internRetry = kwargs.get("retry",  self.retry)
        self._checkingDate()
        if self.loggingEnabled:
            self.logger.info(f"endpoint: {endpoint}")
            self.logger.info(f"params: {params}")
        if headers is None:
            headers = self.header
        if params is None and data is None:
            res = requests.get(
                endpoint, headers=headers)
        elif params is not None and data is None:
            res = requests.get(
                endpoint, headers=headers, params=params)
        elif params is None and data is not None:
            res = requests.get(
                endpoint, headers=headers, data=data)
        elif params is not None and data is not None:
            res = requests.get(endpoint, headers=headers, params=params, data=data)
        if kwargs.get("verbose", False):
            print(f"request URL : {res.request.url}")
            print(f"statut_code : {res.status_code}")
        try:
            while str(res.status_code) == "429":
                if kwargs.get("verbose", False):
                    print(f'Too many requests: retrying in {self.restTime} seconds')
                if self.loggingEnabled:
                    self.logger.info(f"Too many requests: retrying in {self.restTime} seconds")
                time.sleep(self.restTime)
                res = requests.get(endpoint, headers=headers, params=params, data=data)
            res_json = res.json()
        except:
            ## handling 1.4 and classification
            if self.loggingEnabled:
                self.logger.warning(f"handling exception as res.json() cannot be managed")
                self.logger.warning(f"status code: {res.status_code}")
            if kwargs.get('classFile'): ## reading classification files
                text = res.text
                textIO = io.StringIO(text)
                list_lines = list(textIO.readlines())
                data = []
                for line in list_lines:
                    data.append(json.loads(line.strip()))
                return data
            elif kwargs.get('legacy',False): ## handling 1.4
                try:
                    return json.loads(res.text)
                except:
                    if self.loggingEnabled:
                        self.logger.error(f"GET method failed: {res.status_code}, {res.text}")
                    return res.text
            else:
                if self.loggingEnabled:
                    self.logger.error(f"text: {res.text}")
            res_json = {'error': 'Request Error'}
            while internRetry > 0:
                if self.loggingEnabled:
                    self.logger.warning(f"Trying again with internal retry")
                if kwargs.get("verbose", False):
                    print('Retry parameter activated')
                    print(f'{internRetry} retry left')
                if 'error' in res_json.keys():
                    time.sleep(self.restTime)
                    kwargs["retry"] = internRetry - 1
                    res_json = self.getData(endpoint, params=params, data=data, headers=headers, **kwargs)
                    return res_json
        return res_json

    def postData(self, endpoint: str, params: dict = None, data: dict = None, headers: dict = None, files:dict=None, *args, **kwargs):
        """
        Abstraction for posting data
        """
        self._checkingDate()
        if params is None:
            params = {}
        if headers is None:
            headers = self.header
        if data is None and files is None:
            res = requests.post(endpoint, headers=headers,params=params)
        elif data is not None and files is None:
            res = requests.post(endpoint, headers=headers, data=json.dumps(data),params=params)
        elif data is None and files is not None:
            res = requests.post(endpoint, headers=headers, params=params, files=files)
        elif data is not None and files is not None:
            res = requests.post(endpoint, headers=headers, params=params, data=json.dumps(data), files=files)
        try:
            res_json = res.json()
            if res.status_code == 429 or res_json.get('error_code', None) == "429050":
                res_json['status_code'] = 429
        except:
            ## handling 1.4
            if kwargs.get('legacy',False):
                try:
                    return json.loads(res.text)
                except:
                    if self.loggingEnabled:
                        self.logger.error(f"POST method failed: {res.status_code}, {res.text}")
                    return res.text
            if type(res) == dict:
                res_json = {'error': res.get('status_code','Request Error')}
            else:
                res_json = res
        return res_json

    def patchData(self, endpoint: str, params: dict = None, data:dict = None, headers: dict = None, files : dict = None,  *args, **kwargs):
        """
        Abstraction for patching data
        """
        self._checkingDate()
        if headers is None:
            headers = self.header
        if params is not None and data is None and files is None:
            res = requests.patch(endpoint, headers=headers, params=params)
        elif params is None and data is not None and files is None:
            res = requests.patch(endpoint, headers=headers, data=json.dumps(data))
        elif params is not None and data is not None and files is None:
            res = requests.patch(endpoint, headers=headers, params=params, data=json.dumps(data))
        elif files is not None:
            res = requests.patch(endpoint, headers=headers, params=params, files=files)
        try:
            while str(res.status_code) == "429":
                if kwargs.get("verbose", False):
                    print(f'Too many requests: retrying in {self.restTime} seconds')
                time.sleep(self.restTime)
                res = requests.patch(endpoint, headers=headers, params=params,data=json.dumps(data))
            res_json = res.json()
        except:
            if self.loggingEnabled:
                self.logger.error(f"PATCH method failed: {res.status_code}, {res.text}")
            res_json = {'error': res.get('status_code','Request Error')}
        return res_json

    def putData(self, endpoint: str, params: dict = None, data=None, headers: dict = None, files:dict=None,*args, **kwargs):
        """
        Abstraction for putting data
        """
        self._checkingDate()
        if headers is None:
            headers = self.header
        if params is not None and data is None and files is None:
            res = requests.put(endpoint, headers=headers, params=params)
        elif params is None and data is not None and files is None:
            res = requests.put(endpoint, headers=headers, data=json.dumps(data))
        elif params is not None and data is not None and files is None:
            res = requests.put(endpoint, headers=headers, params=params, data=json.dumps(data=data))
        elif files is not None:
            res = requests.put(endpoint, headers=headers, params=params, files=files)
        try:
            status_code = res.json()
        except:
            if self.loggingEnabled:
                self.logger.error(f"PUT method failed: {res.status_code}, {res.text}")
            status_code = {'error': res.get('status_code','Request Error')}
        return status_code

    def deleteData(self, endpoint: str, params: dict = None, headers: dict = None, *args, **kwargs):
        """
        Abstraction for deleting data
        """
        self._checkingDate()
        if headers is None:
            headers = self.header
        if params is None:
            res = requests.delete(endpoint, headers=headers)
        elif params is not None:
            res = requests.delete(endpoint, headers=headers, params=params)
        try:
            while str(res.status_code) == "429":
                if kwargs.get("verbose", False):
                    print(f'Too many requests: retrying in {self.restTime} seconds')
                time.sleep(self.restTime)
                res = requests.delete(endpoint, headers=headers, params=params)
            status_code = res.status_code
        except:
            if self.loggingEnabled:
                self.logger.error(f"DELETE method failed: {res.status_code}, {res.text}")
            status_code = {'error': 'Request Error'}
        return status_code
